/* -*- Mode:C++; c++-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Marco Guastella alias Vasta 
 * Web page:<www.ragnu.it> 
 * Email:<vasta@ragnu.it>
   Date last update: 29/11/2022
*/
 
#include "xml.h"

namespace vampiria { namespace xml {

void dom_parser_error(vmp_uint line,vmp::str error)
{
    vmp::str err;
    vmp::unicode::str_write(&err,"xml::DomParser error line %u: %s",line,error.c_str());
    vmp::except_s(err);
}

Tag::Tag()
{
    init(); 
}

Tag::~Tag()
{
}

void Tag::init()
{
    filepath_="";
    line_=0;
    name_="";
    text_="";
    parent_=0;
    priority_=0;
    order_=0;
}

void Tag::destroy_impl()
{
}

void Tag::set_filepath(vmp::str filepath)
{
     filepath_=filepath;
}
        
vmp::str Tag::get_filepath()
{
    return filepath_;
}

void Tag::set_line(vmp_uint line)
{
    line_=line;
}
        
vmp_uint Tag::get_line()
{
    return line_;
}

void Tag::set_name(vmp::str name)
{
    name_=name;
}
        
vmp::str Tag::get_name()
{
    return name_;
}

void Tag::set_text(vmp::str text)
{
    text_=text;
}
        
vmp::str Tag::get_text()
{
    return text_;
}

void Tag::add_attr(vmp::str name,vmp::str value)
{
    try
    {
        attrs_.insert(name,value);    
    }
    catch(vmp::exception &x)
    {
	vmp::except("net::Tag::add_attr() error name '%s' already",name.c_str());    
    }
}

vmp::str Tag::get_attr(vmp::str name)
{
    vmp::str ret;
    if(attrs_.search(name,&ret))
        return ret;
    return "";
}

vmp::vector<vmp::str> Tag::get_attrs_name()
{
    return attrs_.all_keys();
}

void Tag::set_parent(Tag *parent)
{
    parent_=parent;
}
        
Tag *Tag::get_parent()
{
    return parent_;
}

void Tag::add_child(Tag *child)
{
    childs_.push_back(child);
}
        
Tag *Tag::get_child(vmp_index index)
{
    if(index < childs_.size())
        return childs_[index];
    return 0;
}

vmp_size Tag::get_child_number()
{
    return childs_.size();
}

void Tag::set_priority(vmp_uint priority)
{
    priority_=priority;
}
        
vmp_uint Tag::get_priority()
{
   return priority_;
}

        
void Tag::set_order(vmp_uint order)
{
    order_=order;
}

vmp_uint Tag::get_order()
{
    return order_;
}

void Tag::parser_error(const vmp_char *fmt,...)
{
    va_list ap;
    va_start(ap,fmt);    
    vmp::str sfmt=vmp::va_wrap(fmt,ap);
    va_end(ap);
    parser_error_s(sfmt); 
}

void Tag::parser_error_s(vmp::str error)
{
    vmp::except("%s::%u tag <%s> parser error:\n%s",get_filepath().c_str(),get_line(),get_name().c_str(),error.c_str());
}

void Tag::destroy()
{
    destroy_impl();
    attrs_.clear();
    childs_.clear();
    init();
}

void Tag::build()
{
}

BaseTag::BaseTag():Tag()
{
}
        
BaseTag::~BaseTag()
{
}

PriorityTag::PriorityTag()
{
}

PriorityTag::~PriorityTag()
{
}

vmp_bool PriorityTag::operator()(xml::Tag *&t1,xml::Tag *&t2)
{
    vmp_uint t1p=t1->get_priority();
    vmp_uint t2p=t2->get_priority();
    if((t1p < t2p) || ((t1p == t2p) && (t1->get_order() > t2->get_order())))
        return true;
    return false;
}

DomParser::DomParser()
{
    init();
}

DomParser::~DomParser()
{
}

xml::Tag *DomParser::evaluate(const xmlpp::Node *node,xml::Tag *parent,vmp::str filexml)
{
    xml::Tag *ret=0;
    const xmlpp::ContentNode* nodeContent = dynamic_cast<const xmlpp::ContentNode*>(node);
    const xmlpp::TextNode* nodeText = dynamic_cast<const xmlpp::TextNode*>(node);
    const xmlpp::CommentNode* nodeComment = dynamic_cast<const xmlpp::CommentNode*>(node);

    if(nodeText && !(nodeText->is_white_space()) && parent != 0)
        parent->set_text(vmp::unicode::str_format(nodeText->get_content()));
    Glib::ustring nodename;
    if(!nodeText && !nodeComment && !nodeContent)
        nodename = node->get_name();
    if(!nodeText && !nodeComment && !nodename.empty() && !nodeContent)
    {
        ret=create_tag(nodename);
        ret->set_line(node->get_line());
        ret->set_name(nodename);
        ret->set_order(gorder_++);
        ret->set_filepath(filexml);
        if(parent != 0)
        {
            ret->set_parent(parent);
            parent->add_child(ret);
        }
        const xmlpp::Element* nodeElement = dynamic_cast<const xmlpp::Element*>(node);
        xmlpp::Element::AttributeList attrs=nodeElement->get_attributes();
        for (xmlpp::Element::AttributeList::iterator iter=attrs.begin(); iter !=attrs.end(); ++iter)
        {
             const xmlpp::Attribute* attr=(const xmlpp::Attribute*) (*iter);
             ret->add_attr(attr->get_name(),attr->get_value());
        }
        ptag_.push(ret); //code element
        xmlpp::Node::NodeList list = node->get_children();
        for(xmlpp::Node::NodeList::iterator iter = list.begin(); iter != list.end(); ++iter)
            evaluate(*iter,ret,filexml);
    } 
    return ret;
}

void DomParser::destroy_impl()
{
}

void DomParser::destroy_tag(xml::Tag *tag)
{
    vmp_index i=0;
    xml::Tag *tmp;
    if(tag == 0)
       return;
    while((tmp=tag->get_child(i++)) != 0)
       destroy_tag(tmp);
    tag->destroy();
    delete tag;
}

void DomParser::init()
{
    gorder_=0;
    root_=0;
}

xml::Tag *DomParser::get_root()
{
    return root_;
}

void DomParser::build_impl(xmlpp::DomParser *parser,vmp::str filexml,vmp::str dtd)
{
    xmlpp::Document *doc;
    xmlpp::DtdValidator validator;
    if((doc=parser->get_document()) == 0)
        vmp::except_s("Null parser");
    if(dtd != "")
    {    
        validator.parse_file(dtd);
        validator.validate(doc);
    }
    const xmlpp::Node *rnode=doc->get_root_node();
    reset();
    root_=evaluate(rnode,0,filexml);
    xml::Tag *tmp;        
    while(!ptag_.empty())
    {
        tmp=ptag_.top();
        ptag_.pop();
        tmp->build();
    }
}

void DomParser::build(vmp::str filexml,vmp::str dtd)
{
    xmlpp::DomParser parser;
    try
    {
        parser.set_throw_messages();
        parser.set_substitute_entities();
        parser.parse_file(filexml.c_str());
        build_impl(&parser,filexml,dtd);
    }
    catch(const vmp::exception &x)
    {        
        vmp::except("xml::DomParser::build() %s",x.what());
    }
}

void DomParser::build_from_str(vmp::str xml,vmp::str dtd)
{
    xmlpp::DomParser parser;
    try
    {
        parser.set_throw_messages();
        parser.set_substitute_entities();
        parser.parse_memory(xml.c_str());
        build_impl(&parser,"",dtd);
    }
    catch(const vmp::exception &x)
    {        
        vmp::except("xml::DomParser::build_from_str() %s",x.what());
    }
}

void DomParser::reset()
{
    while(!ptag_.empty())
        ptag_.pop();
    destroy_tag(root_);
    init();  
}

void DomParser::destroy()
{
    reset();
    destroy_impl();
}

WTag::WTag(vmp::str name,WTag *parent)
{
    set_name(name);
    parent_=parent;
    text_="";
}
        
WTag::~WTag()
{
    name_="";
    text_="";
    attrs_.clear();
    for(vmp_index i=0;i<childs_.size();i++)
        delete childs_[i];
    childs_.clear(); 
}

void WTag::write(vmp::str *data,vmp_uint space)
{
    vmp::str attrvalue="";
    vmp::vector<vmp::pair<vmp::str,vmp::str>> values=attrs_.all_pair_values();
    for(vmp_index i=0;i<values.size();i++)
        vmp::unicode::str_cwrite(&attrvalue," %s=\"%s\"",values[i].first.c_str(),values[i].second.c_str());
    vmp::str spacedata="";
    for(vmp_index i=0;i<space;i++)
        vmp::unicode::str_cwrite(&spacedata," ");
    if(childs_.size() == 0)
    {
        if(text_=="")
            vmp::unicode::str_cwrite(data,"%s<%s%s/>\n",spacedata.c_str(),name_.c_str(),attrvalue.c_str());
        else
            vmp::unicode::str_cwrite(data,"%s<%s%s>%s</%s>\n",spacedata.c_str(),name_.c_str(),attrvalue.c_str(),text_.c_str(),name_.c_str());    
    }
    else
    {
        vmp::str datachild="";
        for(vmp_index i=0;i<childs_.size();i++)
            childs_[i]->write(&datachild,space+2);
        vmp::unicode::str_cwrite(data,"%s<%s%s>\n%s%s</%s>\n",spacedata.c_str(),name_.c_str(),attrvalue.c_str(),datachild.c_str(),spacedata.c_str(),name_.c_str());
    }
}

void WTag::copy_tag(xml::Tag *tag)
{
    set_text(tag->get_text());
    vmp::vector<vmp::str> attrs=tag->get_attrs_name();
    for(vmp_index i=0;i<attrs.size();i++)
        set_attr(attrs[i],tag->get_attr(attrs[i]));
    xml::Tag *child;
    WTag *wtag;
    for(vmp_index i=0;i<tag->get_child_number();i++)
    {
        child=tag->get_child(i);
        wtag=new WTag(child->get_name(),this);
        childs_.push_back(wtag);
        wtag->copy_tag(child);
    }
}

void WTag::copy_wtag(WTag *wtag)
{
    set_text(wtag->get_text());
    vmp::vector<vmp::str> attrs=wtag->get_attrs_name();
    for(vmp_index i=0;i<attrs.size();i++)
        set_attr(attrs[i],wtag->get_attr(attrs[i]));
    xml::WTag *child;
    WTag *tmp;
    for(vmp_index i=0;i<wtag->get_child_number();i++)
    {
        child=wtag->get_child(i);
        tmp=new WTag(child->get_name(),this);
        childs_.push_back(tmp);
        tmp->copy_wtag(child);
    }
}

void WTag::set_child_swap(vmp_index i)
{
    WTag *tmp;
    for(vmp_index j=get_child_number()-1;j >i;j--)
    {
        tmp=childs_[j];
        childs_[j]=childs_[j-1];
        childs_[j-1]=tmp;
    }
}

void WTag::set_name(vmp::str tagname)
{
    if(!vmp::unicode::str_isword(tagname))
        vmp::except("xml::WTag::set_name(tagname=%s) bad value",tagname.c_str());    
    name_=tagname;
}

vmp::str WTag::get_name()
{
    return name_;
}

void WTag::set_text(vmp::str text)
{
    text_=vmp::unicode::str_format(text);
}

vmp::str WTag::get_text()
{
    return text_;
}

void WTag::set_attr(vmp::str name,vmp::str value)
{
    if(!vmp::unicode::str_isword(name))
        vmp::except("xml::WTag::set_attr(name=%s) bad value",name.c_str());
    if(!vmp::unicode::str_isword(value))
        vmp::except("xml::WTag::set_attr(value=%s) bad value",value.c_str());
    vmp::str ret;
    try
    {
        attrs_.cancel(name,&ret);
    }
    catch(vmp::exception &x)
    {
    }
    attrs_.insert(name,value);
}

vmp::vector<vmp::str> WTag::get_attrs_name()
{
    return attrs_.all_keys();
}

vmp::str WTag::get_attr(vmp::str name)
{
    if(!vmp::unicode::str_isword(name))
        vmp::except("xml::WTag::get_attr(name=%s) bad value",name.c_str());
    vmp::str ret;
    if(!attrs_.search(name,&ret))
        vmp::except("xml::WTag::get_attr(name=%s) attribute not found",name.c_str());
    return ret;
}

void WTag::del_attr(vmp::str name)
{
    if(!vmp::unicode::str_isword(name))
        vmp::except("xml::WTag::del_attr(name=%s) bad value",name.c_str());
    vmp::str ret;
    try
    {
        attrs_.cancel(name,&ret);
    }
    catch(vmp::exception &x)
    {
        vmp::except("xml::WTag::del_attr(name=%s) attribute not found",name.c_str());
    }
}

WTag *WTag::add_child(vmp::str name)
{
    WTag *child=new WTag(name,this);
    childs_.push_back(child);
    return child;
}

WTag *WTag::add_child_from_tag(xml::Tag *tag)
{
    vmp::except_check_pointer((void *) tag,"xml::WTag::add_child_from_tag(tag=NULL)");   
    WTag *child=new xml::WTag(tag->get_name());
    child->copy_tag(tag);
    childs_.push_back(child);   
    return child;
}

WTag *WTag::add_child_from_wtag(WTag *wtag)
{
    vmp::except_check_pointer((void *) wtag,"xml::WTag::add_child_from_wtag(wtag=NULL)");   
    WTag *child=new xml::WTag(wtag->get_name());
    child->copy_wtag(wtag);
    childs_.push_back(child);   
    return child;
}

WTag *WTag::set_child(vmp::str name,vmp_index i)
{
    if(get_child_number() <= i)
        vmp::except("xml::Wtag::set_child(i=%d) out of range",i);
    WTag *child=add_child(name);
    set_child_swap(i);
    return child;
}

WTag *WTag::set_child_from_tag(xml::Tag *tag,vmp_index i)
{
    vmp::except_check_pointer((void *) tag,"xml::WTag::set_child_from_tag(tag=NULL)"); 
    if(get_child_number() <= i)
        vmp::except("xml::Wtag::set_child_from_tag(i=%d) out of range",i);
    WTag *child=add_child_from_tag(tag);
    set_child_swap(i);
    return child;
}

WTag *WTag::set_child_from_wtag(WTag *wtag,vmp_index i)
{
    vmp::except_check_pointer((void *) wtag,"xml::WTag::set_child_from_wtag(wtag=NULL)");
    if(get_child_number() <= i)
        vmp::except("xml::Wtag::set_child_from_tag(i=%d) out of range",i);
    WTag *child=add_child_from_wtag(wtag);
    set_child_swap(i);
    return child;
}

WTag *WTag::replace_child(vmp::str name,vmp_index i)
{
    vmp_size size=get_child_number();
    if(size <= i)
        vmp::except("xml::Wtag::replace_child(i=%d) out of range",i);
    WTag *child=new WTag(name,this);
    delete childs_[i];
    childs_[i]=child;
    return child;
}

WTag *WTag::replace_child_from_tag(xml::Tag *tag,vmp_index i)
{
    vmp_size size=get_child_number();
    if(size <= i)
        vmp::except("xml::Wtag::replace_child_from_tag(i=%d) out of range",i);
    vmp::except_check_pointer((void *) tag,"xml::WTag::replace_child_from_tag(tag=NULL)");
    WTag *child=new WTag(tag->get_name(),this);
    child->copy_tag(tag);
    delete childs_[i];
    childs_[i]=child;
    return child;
}

WTag *WTag::replace_child_from_wtag(WTag *wtag,vmp_index i)
{
    vmp_size size=get_child_number();
    if(size <= i)
        vmp::except("xml::Wtag::replace_child_from_wtag(i=%d) out of range",i);
    vmp::except_check_pointer((void *) wtag,"xml::WTag::replace_child_from_wtag(tag=NULL)");
    WTag *child=new WTag(wtag->get_name(),this);
    child->copy_wtag(wtag);
    delete childs_[i];
    childs_[i]=child;
    return child;
}

vmp_size WTag::get_child_number()
{
    return childs_.size();
}

WTag *WTag::get_child(vmp_index i)
{
    if(get_child_number() <= i)
        return 0;
    return childs_[i];
}

vmp_size WTag::del_child(vmp_index i)
{
    vmp_size size=get_child_number();
    if(size <= i)
        vmp::except("xml::Wtag::del_child(i=%d) out of range",i);
    delete childs_[i];
    for(vmp_index j=i+1;j<size;j++)
        childs_[j-1]=childs_[j];
    childs_.pop_back();
    return get_child_number();
}

WTag *WTag::get_parent()
{
    return parent_;
}

WParser::WParser()
{
    root_=0;
}
        
WParser::~WParser()
{
    reset();
}

xml::WTag *WParser::get_root()
{
    vmp::except_check_pointer((void *) root_,"xml::WParser::get_root() null parser");
    return root_;
}

vmp::str WParser::write_impl()
{
    vmp::str data;
    vmp::unicode::str_write(&data,"%s\n",vmp::env::xml_hdr().c_str());
    root_->write(&data,0);
    return data;
}

void WParser::reset()
{
    if(root_ != 0)
    {
        delete root_;
        root_=0;
    }
}

xml::WTag *WParser::init(vmp::str rootname)
{
    reset();
    root_=new WTag(rootname);
    return root_; 
}

xml::WTag *WParser::init_from_tag(xml::Tag *tag)
{
    vmp::except_check_pointer((void *) tag,"xml::WParser::init_from_tag(tag=NULL)");   
    reset();
    root_=new xml::WTag(tag->get_name());
    root_->copy_tag(tag);
    return root_;
}

xml::WTag *WParser::init_from_wtag(xml::WTag *wtag)
{
    vmp::except_check_pointer((void *) wtag,"xml::WParser::init_from_wtag(wtag=NULL)");   
    reset();
    root_=new xml::WTag(wtag->get_name());
    root_->copy_wtag(wtag);
    return root_;
}

vmp::str WParser::write_to_str()
{
    vmp::str ret;
    vmp::except_check_pointer((void *) root_,"xml::WParser::write_to_str() undefined parser");
    try
    {
        ret=write_impl();
    }
    catch(vmp::exception &x)
    {
        vmp::except("xml::WParser::write_to_str() error '%s'",x.what());   
    }
    return ret;
}
void WParser::write(vmp::str filepath)
{
    vmp::str ret;
    vmp::except_check_pointer((void *) root_,"xml::WParser::write() undefined parser");
    try
    {
        ret=write_impl();
        vmp::output_file(filepath,true,"%s",ret.c_str());
    }
    catch(vmp::exception &x)
    {
        vmp::except("xml::WParser::write() error '%s'",x.what());   
    }
}

}}
 
